# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import ipaddress
from abc import ABC, abstractmethod
from someipy._internal.someip_sd_header import (
    SdService,
    SdEventGroupEntry,
    SdIPV4EndpointOption,
)
from someipy._internal.session_handler import SessionHandler


class ServiceDiscoveryObserver(ABC):
    @abstractmethod
    def handle_offer_service(self, offered_service: SdService) -> None:
        pass

    @abstractmethod
    def handle_stop_offer_service(self, offered_service: SdService) -> None:
        pass

    @abstractmethod
    def handle_find_service(self) -> None:
        pass

    @abstractmethod
    def handle_subscribe_eventgroup(
        self,
        sd_event_group: SdEventGroupEntry,
        ip4_endpoint_option: SdIPV4EndpointOption,
    ) -> None:
        pass

    @abstractmethod
    def handle_subscribe_ack_eventgroup(
        self, sd_event_group: SdEventGroupEntry
    ) -> None:
        pass


class ServiceDiscoverySubject(ABC):
    @abstractmethod
    def attach(self, service_instance: ServiceDiscoveryObserver) -> None:
        pass

    @abstractmethod
    def detach(self, service_instance: ServiceDiscoveryObserver) -> None:
        pass


class ServiceDiscoverySender(ABC):

    @abstractmethod
    def offer_service(self, service: SdService) -> None:
        pass

    @abstractmethod
    def send_multicast(self, buffer: bytes) -> None:
        pass

    @abstractmethod
    def send_unicast(self, buffer: bytes, dest_ip: ipaddress.IPv4Address) -> None:
        pass

    @abstractmethod
    def get_multicast_session_handler(self) -> SessionHandler:
        pass

    @abstractmethod
    def get_unicast_session_handler(self) -> SessionHandler:
        pass



# Copyright (C) 2024 Christian H.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from abc import ABC, abstractmethod
from typing import Callable, Tuple, Any, Union
from someipy._internal.someip_message import SomeIpMessage
from someipy._internal.tcp_client_manager import (
    TcpClientManagerInterface,
    TcpClientProtocolInterface,
)
from someipy._internal.utils import EndpointType
from someipy._internal.someip_data_processor import SomeipDataProcessor


class SomeipEndpoint(ABC):
    """
    A class representing an endpoint (UDP or TCP) which fires a callback when a SomeIpMessage is received.
    It can also send messages to a specific endpoint or broadcast to all connected endpoints (for UDP multicast or to all connected
    TCP clients).
    """

    @abstractmethod
    def set_someip_callback(
        self, callback_func: Callable[[SomeIpMessage, Tuple[str, int]], None]
    ) -> None:
        pass

    @abstractmethod
    def sendto(self, data: bytes, addr: EndpointType) -> None:
        pass

    @abstractmethod
    def sendtoall(self, data: bytes) -> None:
        pass

    @abstractmethod
    def shutdown(self) -> None:
        pass


class UDPSomeipEndpoint(SomeipEndpoint, asyncio.DatagramProtocol):
    def __init__(self):
        self._callback: Callable[[SomeIpMessage, Tuple[str, int]], None] = None
        self._transport = None
        self._processor = SomeipDataProcessor()

    def set_someip_callback(
        self, callback_func: Callable[[SomeIpMessage, Tuple[str, int]], None]
    ) -> None:
        self._callback = callback_func

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
        self._transport = transport

    def connection_lost(self, exc: Exception) -> None:
        pass

    def datagram_received(self, data: bytes, addr: Tuple[Union[str, Any, int]]) -> None:
        result = self._processor.process_data(data)
        if result and self._callback is not None:
            self._callback(self._processor.someip_message, addr)

    def sendto(self, data: bytes, addr: EndpointType) -> None:
        if self._transport is not None:
            self._transport.sendto(data, addr)

    def sendtoall(self, data: bytes) -> None:
        # TODO: Implement for multicast support
        raise NotImplementedError("No implementation for UDP yet")

    def shutdown(self) -> None:
        if self._transport is not None:
            self._transport.close()


class TCPSomeipEndpoint(SomeipEndpoint):
    def __init__(self, server, manager):
        self._server: asyncio.Server = server
        self._manager: TcpClientManagerInterface = manager

    def set_someip_callback(
        self, callback_func: Callable[[SomeIpMessage, Tuple[str, int]], None]
    ) -> None:
        self._manager.register_callback(callback_func)

    def sendto(self, data: bytes, addr: EndpointType) -> None:
        if self._manager is None:
            return
        client: TcpClientProtocolInterface = self._manager.get_client(addr[0], addr[1])
        if client is not None:
            client.write(data)

    def sendtoall(self, data: bytes) -> None:
        if self._manager is None:
            return
        for c in self._manager.get_all_clients():
            c.write(data)

    def shutdown(self) -> None:
        if self._manager is None:
            return
        self._server.close()



from dataclasses import dataclass
from someipy.serialization import (
    Uint8,
    Uint16,
    Uint32,
    Uint64,
    Sint8,
    Sint16,
    Sint32,
    Sint64,
    Bool,
    Float32,
    Float64,
    SomeIpPayload,
    SomeIpFixedSizeArray,
)


def test_base_types_len():
    assert 1 == len(Uint8(1))
    assert 2 == len(Uint16(1))
    assert 4 == len(Uint32(1))
    assert 8 == len(Uint64(1))

    assert 1 == len(Sint8(-1))
    assert 2 == len(Sint16(-1))
    assert 4 == len(Sint32(-1))
    assert 8 == len(Sint64(-1))

    assert 1 == len(Bool(True))
    assert 4 == len(Float32(1.0))
    assert 8 == len(Float64(1.0))


@dataclass
class MsgBaseTypesOnly(SomeIpPayload):
    """
    The dataclass decorator will generate an __eq__ method which
    can be used for comparing the content of two messages
    """

    # Expected 13 bytes in total
    x: Uint8  # 1 byte
    y: Uint32  # 4 bytes
    z: Float64  # 8 bytes

    def __init__(self):
        self.x = Uint8(0)
        self.y = Uint32(0)
        self.z = Float64(0.0)


@dataclass
class MsgWithOneStruct(SomeIpPayload):
    # Expected 1 + 13 + 4 bytes in total
    a: Uint8
    b: MsgBaseTypesOnly
    c: Sint32

    def __init__(self):
        self.a = Uint8(0)
        self.b = MsgBaseTypesOnly()
        self.c = Sint32(0)


def test_someip_payload_len():
    assert 13 == len(MsgBaseTypesOnly())
    assert (13 + 1 + 4) == len(MsgWithOneStruct())


def test_struct_equals_operator():
    m_1 = MsgBaseTypesOnly()
    m_2 = MsgBaseTypesOnly()
    assert m_1 == m_2

    m_1.x = Uint8(1)
    assert m_1 != m_2

    m_2.x = Uint8(1)
    assert m_1 == m_2


def test_struct_serialization_and_deserialization():
    m = MsgBaseTypesOnly()
    m.y = Uint32(4)
    # Expected: 0x 00 00000004 0000000000000000
    assert bytes.fromhex("00000000040000000000000000") == m.serialize()

    m.x = Uint8(255)
    # Expected: 0x FF 00000004 0000000000000000
    assert bytes.fromhex("ff000000040000000000000000") == m.serialize()

    m_again = MsgBaseTypesOnly().deserialize(
        bytes.fromhex("ff000000040000000000000000")
    )
    assert m_again == m

    # The first byte is changed now to decimal 254
    m_different = MsgBaseTypesOnly().deserialize(
        bytes.fromhex("fe000000040000000000000000")
    )
    assert m_different != m_again
    assert m_different != m


def test_fixed_size_array_length():
    a = SomeIpFixedSizeArray(Uint16, 10)
    expected_bytes = len(Uint16()) * 10
    assert len(a) == expected_bytes


def test_fixed_size_array_equals_operator():
    a = SomeIpFixedSizeArray(Uint16, 10)
    b = SomeIpFixedSizeArray(Uint16, 10)
    assert a == b

    c = SomeIpFixedSizeArray(Uint8, 10)
    assert a != c

    a.data[0] = Uint16(5)
    assert a != b

    b.data[0] = Uint16(5)
    assert a == b


def test_fixed_size_array_serialization_and_deserialization():
    a = SomeIpFixedSizeArray(Uint8, 4)
    # Expected hex: 0x 00 00 00 00
    assert bytes.fromhex("00000000") == a.serialize()

    a.data[0] = Uint8(1)
    a.data[1] = Uint8(2)
    a.data[2] = Uint8(3)
    a.data[3] = Uint8(4)
    # Expected hex: 0x 01 02 03 04
    assert bytes.fromhex("01020304") == a.serialize()

    a_again = SomeIpFixedSizeArray(Uint8, 4).deserialize(bytes.fromhex("01020304"))
    assert a_again == a



from ._internal.transport_layer_protocol import TransportLayerProtocol  # noqa: F401
from .service import Service, ServiceBuilder, EventGroup  # noqa: F401
from .server_service_instance import (
    ServerServiceInstance,
    construct_server_service_instance,
)  # noqa: F401
from .client_service_instance import (
    ClientServiceInstance,
    construct_client_service_instance,
    MethodResult,
)  # noqa: F401

from ._internal.someip_message import SomeIpMessage  # noqa: F401
from ._internal.method_result import MethodResult  # noqa: F401
from ._internal.return_codes import ReturnCode  # noqa: F401
from ._internal.message_types import MessageType  # noqa: F401



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from enum import Enum
import struct
from typing import Iterable, Tuple, Callable, Set, List

from someipy import Service
from someipy._internal.method_result import MethodResult
from someipy._internal.someip_data_processor import SomeipDataProcessor
from someipy._internal.someip_sd_header import (
    SdService,
    TransportLayerProtocol,
    SdEventGroupEntry,
)
from someipy._internal.someip_header import (
    SomeIpHeader,
)
from someipy._internal.someip_sd_builder import build_subscribe_eventgroup_sd_header
from someipy._internal.service_discovery_abcs import (
    ServiceDiscoveryObserver,
    ServiceDiscoverySender,
)
from someipy._internal.utils import (
    create_udp_socket,
    EndpointType,
    endpoint_to_str_int_tuple,
)
from someipy._internal.logging import get_logger
from someipy._internal.message_types import MessageType
from someipy._internal.return_codes import ReturnCode
from someipy._internal.someip_endpoint import (
    SomeipEndpoint,
    UDPSomeipEndpoint,
    SomeIpMessage,
)
from someipy._internal.tcp_connection import TcpConnection

_logger_name = "client_service_instance"


class ExpectedAck:
    def __init__(self, eventgroup_id: int) -> None:
        self.eventgroup_id = eventgroup_id

    def __eq__(self, value: object) -> bool:
        return self.eventgroup_id == value.eventgroup_id


class FoundService:
    service: SdService

    def __init__(self, service: SdService) -> None:
        self.service = service

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, FoundService):
            return self.service == __value.service


class ClientServiceInstance(ServiceDiscoveryObserver):
    _service: Service
    _instance_id: int
    _endpoint: EndpointType
    _protocol: TransportLayerProtocol
    _someip_endpoint: SomeipEndpoint
    _ttl: int
    _sd_sender: ServiceDiscoverySender

    _eventgroups_to_subscribe: Set[int]
    _expected_acks: List[ExpectedAck]

    _callback: Callable[[bytes], None]
    _found_services: Iterable[FoundService]
    _subscription_active: bool
    _method_call_future: asyncio.Future

    def __init__(
        self,
        service: Service,
        instance_id: int,
        endpoint: EndpointType,
        protocol: TransportLayerProtocol,
        someip_endpoint: SomeipEndpoint,
        ttl: int = 0,
        sd_sender=None,
    ):
        self._service = service
        self._instance_id = instance_id
        self._endpoint = endpoint
        self._protocol = protocol
        self._someip_endpoint = someip_endpoint
        self._ttl = ttl
        self._sd_sender = sd_sender

        self._eventgroups_to_subscribe = set()
        self._expected_acks = []
        self._callback = None

        self._tcp_connection: TcpConnection = None

        self._tcp_connect_lock = asyncio.Lock()
        self._tcp_task = None
        self._tcp_connection_established_event = asyncio.Event()
        self._shutdown_requested = False

        self._found_services = []
        self._subscription_active = False
        self._method_call_future = None

    def register_callback(self, callback: Callable[[SomeIpMessage], None]) -> None:
        """
        Register a callback function to be called when a SOME/IP event is received.

        Args:
            callback (Callable[[SomeIpMessage], None]): The callback function to be registered.
                This function should take a SomeIpMessage object as its only argument and return None.

        Returns:
            None
        """
        self._callback = callback

    def service_found(self) -> bool:
        """
        Returns whether the service instance represented by the ClientServiceInstance has been offered by a server and was found.
        """
        has_service = False
        for s in self._found_services:
            if (
                s.service.service_id == self._service.id
                and s.service.instance_id == self._instance_id
            ):
                has_service = True
                break
        return has_service

    async def call_method(self, method_id: int, payload: bytes) -> MethodResult:
        """
        Calls a method on the service instance represented by the ClientServiceInstance.

        Args:
            method_id (int): The ID of the method to call.
            payload (bytes): The payload to send with the method call.

        Returns:
            MethodResult: The result of the method call which can contain an error or a successfull result including the response payload.

        Raises:
            RuntimeError: If the TCP connection to the server cannot be established or if the server service has not been found yet.
            asyncio.TimeoutError: If the method call times out, i.e. the server does not send back a response within one second.
        """

        get_logger(_logger_name).debug(f"Try to call method 0x{method_id:04X}")

        if not self.service_found():
            get_logger(_logger_name).warning(
                f"Method 0x{method_id:04x} called, but service 0x{self._service.id:04X} with instance 0x{self._instance_id:04X} not found yet."
            )
            raise RuntimeError(
                f"Method 0x{method_id:04x} called, but service 0x{self._service.id:04X} with instance 0x{self._instance_id:04X} not found yet."
            )

        header = SomeIpHeader(
            service_id=self._service.id,
            method_id=method_id,
            client_id=0x00,
            session_id=0x00,
            protocol_version=0x01,
            interface_version=0x00,
            message_type=MessageType.REQUEST.value,
            return_code=0x00,
            length=len(payload) + 8,
        )
        someip_message = SomeIpMessage(header, payload)

        self._method_call_future = asyncio.get_running_loop().create_future()

        dst_address = str(self._found_services[0].service.endpoint[0])
        dst_port = self._found_services[0].service.endpoint[1]

        if self._protocol == TransportLayerProtocol.TCP:
            # In case of TCP, first try to connect to the TCP server
            # [PRS_SOMEIP_00708] The TCP connection shall be opened by the client, when the
            # first method call shall be transported or the client tries to receive the first notifications
            if self._tcp_task is None:
                get_logger(_logger_name).debug(
                    f"Create new TCP task for client of 0x{self._instance_id:04X}, 0x{self._service.id:04X}"
                )
                self._tcp_task = asyncio.create_task(
                    self.setup_tcp_connection(
                        str(self._endpoint[0]), self._endpoint[1], dst_address, dst_port
                    )
                )

            try:
                # Wait for two seconds until the connection is established, otherwise return an error
                await asyncio.wait_for(self._tcp_connection_established_event.wait(), 2)
            except asyncio.TimeoutError:
                get_logger(_logger_name).error(
                    f"Cannot establish TCP connection to {dst_address}:{dst_port}."
                )
                raise RuntimeError(
                    f"Cannot establish TCP connection to {dst_address}:{dst_port}."
                )

            if self._tcp_connection.is_open():
                self._tcp_connection.writer.write(someip_message.serialize())
            else:
                get_logger(_logger_name).error(
                    f"TCP connection to {dst_address}:{dst_port} is not opened."
                )
                raise RuntimeError(
                    f"TCP connection to {dst_address}:{dst_port} is not opened."
                )

        else:
            # In case of UDP, just send out the datagram and wait for the response
            self._someip_endpoint.sendto(
                someip_message.serialize(),
                endpoint_to_str_int_tuple(self._found_services[0].service.endpoint),
            )

        # After sending the method call wait for maximum three seconds
        try:
            await asyncio.wait_for(self._method_call_future, 3.0)
        except asyncio.TimeoutError:
            get_logger(_logger_name).error(
                f"Waiting on response for method call 0x{method_id:04X} timed out."
            )
            raise

        return self._method_call_future.result()

    def someip_message_received(
        self, someip_message: SomeIpMessage, addr: Tuple[str, int]
    ) -> None:
        if (
            someip_message.header.client_id == 0x00
            and someip_message.header.message_type == MessageType.NOTIFICATION.value
            and someip_message.header.return_code == ReturnCode.E_OK.value
        ):
            if self._callback is not None and self._subscription_active:
                self._callback(someip_message)
                return

        if (
            someip_message.header.message_type == MessageType.RESPONSE.value
            or someip_message.header.message_type == MessageType.ERROR.value
        ):
            if self._method_call_future is not None:
                result = MethodResult()
                result.message_type = MessageType(someip_message.header.message_type)
                result.return_code = ReturnCode(someip_message.header.return_code)
                result.payload = someip_message.payload
                self._method_call_future.set_result(result)
                return

    def subscribe_eventgroup(self, eventgroup_id: int):
        """
        Adds an event group to the list of event groups to subscribe to.

        Args:
            eventgroup_id (int): The ID of the event group to subscribe to.

        Returns:
            None

        Raises:
            None

        Notes:
            - If the event group ID is already in the subscription list, a debug log message is printed.
        """
        if eventgroup_id in self._eventgroups_to_subscribe:
            get_logger(_logger_name).debug(
                f"Eventgroup ID {eventgroup_id} is already in subscription list."
            )
        self._eventgroups_to_subscribe.add(eventgroup_id)

    def stop_subscribe_eventgroup(self, eventgroup_id: int):
        """
        Stops subscribing to an event group. Not implemented yet.

        Args:
            eventgroup_id (int): The ID of the event group to stop subscribing to.

        Raises:
            NotImplementedError: This method is not yet implemented.

        Notes:
            - This method is currently not implemented and raises a `NotImplementedError`.
        """
        # TODO: Implement StopSubscribe
        raise NotImplementedError

    def handle_find_service(self):
        # Not needed in client service instance
        pass

    def handle_offer_service(self, offered_service: SdService):
        if self._service.id != offered_service.service_id:
            return
        if self._instance_id != offered_service.instance_id:
            return

        if (
            offered_service.service_id == self._service.id
            and offered_service.instance_id == self._instance_id
        ):
            if FoundService(offered_service) not in self._found_services:
                self._found_services.append(FoundService(offered_service))

            if len(self._eventgroups_to_subscribe) == 0:
                return

            # Try to subscribe to requested event groups
            for eventgroup_to_subscribe in self._eventgroups_to_subscribe:
                (
                    session_id,
                    reboot_flag,
                ) = self._sd_sender.get_unicast_session_handler().update_session()

                # Improvement: Pack all entries into a single SD message
                subscribe_sd_header = build_subscribe_eventgroup_sd_header(
                    service_id=self._service.id,
                    instance_id=self._instance_id,
                    major_version=self._service.major_version,
                    ttl=self._ttl,
                    event_group_id=eventgroup_to_subscribe,
                    session_id=session_id,
                    reboot_flag=reboot_flag,
                    endpoint=self._endpoint,
                    protocol=self._protocol,
                )

                get_logger(_logger_name).debug(
                    f"Send subscribe for instance 0x{self._instance_id:04X}, service: 0x{self._service.id:04X}, "
                    f"eventgroup ID: {eventgroup_to_subscribe} TTL: {self._ttl}, version: "
                    f"session ID: {session_id}"
                )

                if self._protocol == TransportLayerProtocol.TCP:
                    if self._tcp_task is None:
                        get_logger(_logger_name).debug(
                            f"Create new TCP task for client of 0x{self._instance_id:04X}, 0x{self._service.id:04X}"
                        )
                        self._tcp_task = asyncio.create_task(
                            self.setup_tcp_connection(
                                str(self._endpoint[0]),
                                self._endpoint[1],
                                str(offered_service.endpoint[0]),
                                offered_service.endpoint[1],
                            )
                        )

                self._expected_acks.append(ExpectedAck(eventgroup_to_subscribe))
                self._sd_sender.send_unicast(
                    buffer=subscribe_sd_header.to_buffer(),
                    dest_ip=offered_service.endpoint[0],
                )

    def handle_stop_offer_service(self, offered_service: SdService) -> None:
        if self._service.id != offered_service.service_id:
            return
        if self._instance_id != offered_service.instance_id:
            return

        # Remove the service from the found services
        self._found_services = [
            f for f in self._found_services if f.service != offered_service
        ]

        self._expected_acks = []
        self._subscription_active = False

    async def setup_tcp_connection(
        self, src_ip: str, src_port: int, dst_ip: str, dst_port: int
    ):
        try:
            while True:

                get_logger(_logger_name).debug(
                    f"Try to open TCP connection to ({dst_ip}, {dst_port})"
                )
                self._tcp_connection = TcpConnection(dst_ip, dst_port)

                # Reset the event before the first await call
                self._tcp_connection_established_event.clear()
                try:
                    await self._tcp_connection.connect(src_ip, src_port)
                except OSError:
                    get_logger(_logger_name).debug(
                        f"Connection refused to ({dst_ip}, {dst_port}). Try to reconnect in 1 second"
                    )
                    # Wait a second before trying to connect again
                    await asyncio.sleep(1.0)
                    continue

                # Notify other tasks waiting on the event, so the other task could send a method call
                if self._tcp_connection.is_open():
                    self._tcp_connection_established_event.set()

                get_logger(_logger_name).debug(f"Start reading on port {src_port}")

                someip_processor = SomeipDataProcessor()

                while self._tcp_connection.is_open():
                    try:
                        new_data = await asyncio.wait_for(
                            self._tcp_connection.reader.read(
                                someip_processor.expected_bytes
                            ),
                            3.0,
                        )

                        if someip_processor.process_data(new_data):
                            self.someip_message_received(
                                someip_processor.someip_message, (dst_ip, dst_port)
                            )

                    except asyncio.TimeoutError:
                        get_logger(_logger_name).debug(
                            f"Timeout reading from TCP connection ({src_ip}, {src_port})"
                        )

                # Clear the event to avoid that a method call would be sent
                self._tcp_connection_established_event.clear()
                await self._tcp_connection.close()

        except asyncio.CancelledError:
            if self._tcp_connection.is_open():
                await self._tcp_connection.close()
            get_logger(_logger_name).debug("TCP task is cancelled. Raise again.")
            raise

    def handle_subscribe_eventgroup(self, _, __) -> None:
        # Not needed for client instance
        pass

    def handle_subscribe_ack_eventgroup(
        self, event_group_entry: SdEventGroupEntry
    ) -> None:
        new_acks: List[ExpectedAck] = []
        ack_found = False
        for expected_ack in self._expected_acks:
            if expected_ack.eventgroup_id == event_group_entry.eventgroup_id:
                ack_found = True
                self._subscription_active = True
                get_logger(_logger_name).debug(
                    f"Received expected subscribe ACK for instance 0x{event_group_entry.sd_entry.instance_id:04X}, service 0x{event_group_entry.sd_entry.service_id:04X}, eventgroup 0x{event_group_entry.eventgroup_id:04X}"
                )
            else:
                new_acks.append(expected_ack)

        self._expected_acks = new_acks
        if not ack_found:
            get_logger(_logger_name).warning(
                f"Received unexpected subscribe ACK for instance 0x{event_group_entry.sd_entry.instance_id:04X}, service 0x{event_group_entry.sd_entry.service_id:04X}, eventgroup 0x{event_group_entry.eventgroup_id:04X}"
            )

    async def close(self):
        self._shutdown_requested = True
        if self._tcp_task is not None:
            self._tcp_task.cancel()
            try:
                await self._tcp_task
            except asyncio.CancelledError:
                get_logger(_logger_name).debug("TCP task is cancelled.")


async def construct_client_service_instance(
    service: Service,
    instance_id: int,
    endpoint: EndpointType,
    ttl: int = 0,
    sd_sender=None,
    protocol=TransportLayerProtocol.UDP,
) -> ClientServiceInstance:
    """
    Asynchronously constructs a ClientServerInstance. Based on the given transport protocol, proper endpoints are setup before constructing the actual ServerServiceInstance.

    Args:
        service (Service): The service associated with the instance.
        instance_id (int): The ID of the instance.
        endpoint (EndpointType): The endpoint of the client instance containing IP address and port.
        ttl (int, optional): The time-to-live for the instance used for service discovery subscribe entries. A value of 0 means that subscriptions are valid for infinite time.
        sd_sender (Any, optional): The service discovery sender.
        protocol (TransportLayerProtocol, optional): The transport layer protocol for the instance. Defaults to TransportLayerProtocol.UDP.

    Returns:
        ClientServerInstance: The constructed ClientServerInstance.

    Raises:
        None
    """
    if protocol == TransportLayerProtocol.UDP:
        loop = asyncio.get_running_loop()
        rcv_socket = create_udp_socket(str(endpoint[0]), endpoint[1])

        _, udp_endpoint = await loop.create_datagram_endpoint(
            lambda: UDPSomeipEndpoint(), sock=rcv_socket
        )

        client_instance = ClientServiceInstance(
            service,
            instance_id,
            endpoint,
            TransportLayerProtocol.UDP,
            udp_endpoint,
            ttl,
            sd_sender,
        )

        udp_endpoint.set_someip_callback(client_instance.someip_message_received)

        return client_instance

    elif protocol == TransportLayerProtocol.TCP:

        client_instance = ClientServiceInstance(
            service,
            instance_id,
            endpoint,
            TransportLayerProtocol.TCP,
            None,
            ttl,
            sd_sender,
        )

        return client_instance

    client_instance = ClientServiceInstance(service, instance_id, ttl, sd_sender)

    return client_instance



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import struct

from dataclasses import dataclass
from typing import Generic, List, Type, TypeVar


"""
PRS_SOMEIP_00065
This table about basic datatypes is taken from
https://www.autosar.org/fileadmin/standards/R22-11/FO/AUTOSAR_PRS_SOMEIPProtocol.pdf

Type Description Size [bit] Remark
boolean TRUE/FALSE value 8 FALSE (0), TRUE (1)
uint8 unsigned Integer 8
uint16 unsigned Integer 16
uint32 unsigned Integer 32
uint64 unsigned Integer 64
sint8 signed Integer 8 
sint16 signed Integer 16
sint32 signed Integer 32
sint64 signed Integer 64
float32 floating point number 32 IEEE 754 binary32 (Single Precision)
float64 floating point number 64 IEEE 754 binary64 (Double Precision)
"""


@dataclass
class Uint8:
    """
    someipy datatype representing an unsigned 8 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        """
        Return the length of the object.

        :return: An integer representing the length of the object.
        :rtype: int
        """
        return 1

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">B", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            None
        """
        (self.value,) = struct.unpack(">B", payload)


@dataclass
class Sint8:
    """
    someipy datatype representing a signed 8 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        """
        Return the length of the object.

        :return: An integer representing the length of the object.
        :rtype: int
        """
        return 1

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">b", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.
        """
        (self.value,) = struct.unpack(">b", payload)
        return self


@dataclass
class Uint16:
    """
    someipy datatype representing an unsigned 16 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 2.
        """
        return 2

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">H", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.
        """
        (self.value,) = struct.unpack(">H", payload)
        return self


@dataclass
class Sint16:
    """
    someipy datatype representing a signed 16 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 2.
        """
        return 2

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">h", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.
        """
        (self.value,) = struct.unpack(">h", payload)
        return self


@dataclass
class Uint32:
    """
    someipy datatype representing an unsigned 32 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        return 4

    def serialize(self) -> bytes:
        return struct.pack(">L", self.value)

    def deserialize(self, payload):
        (self.value,) = struct.unpack(">L", payload)
        return self


@dataclass
class Sint32:
    """
    someipy datatype representing a signed 32 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        return 4

    def serialize(self) -> bytes:
        return struct.pack(">l", self.value)

    def deserialize(self, payload):
        (self.value,) = struct.unpack(">l", payload)
        return self


@dataclass
class Uint64:
    """
    someipy datatype representing an unsigned 64 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        return 8

    def serialize(self) -> bytes:
        return struct.pack(">Q", self.value)

    def deserialize(self, payload):
        (self.value,) = struct.unpack(">Q", payload)
        return self


@dataclass
class Sint64:
    """
    someipy datatype representing a signed 64 bit integer on the wire.
    """

    value: int = 0

    def __len__(self):
        return 8

    def serialize(self) -> bytes:
        return struct.pack(">q", self.value)

    def deserialize(self, payload):
        (self.value,) = struct.unpack(">q", payload)
        return self


@dataclass
class Bool:
    """
    someipy datatype representing a boolean type transported as a single byte on the wire.
    """

    value: bool = False

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 1.
        """
        return 1

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.
        """
        return struct.pack(">B", int(self.value))

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the value of the object. It expects the payload to be a single byte representing a boolean value. If the payload is 0, the value of the object is set to False. If the payload is 1, the value of the object is set to True. The deserialized object is then returned.
        """
        (int_value,) = struct.unpack(">B", payload)
        if int_value == 0:
            self.value = False
        elif int_value == 1:
            self.value = True
        return self


@dataclass
class Float32:
    """
    someipy datatype representing a 32 bit floating type.
    """

    value: float = 0.0

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 4.
        """
        return 4

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.

        This method serializes the value of the object into bytes using the big-endian byte order. It expects the value to be a float. The serialized value is returned as a bytes object.
        """
        return struct.pack(">f", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the value of the object. It expects the payload to be a 4-byte float in big-endian byte order. The deserialized value is assigned to the `value` attribute of the object. The deserialized object is then returned.
        """
        (self.value,) = struct.unpack(">f", payload)
        return self

    def __eq__(self, other) -> Bool:
        """
        Compare two objects for equality.

        This method compares the serialized representation of the current object and the other object to determine if they are equal. It returns a Bool object indicating whether the objects are equal or not.

        Parameters:
            other (Any): The object to compare with the current object.

        Returns:
            Bool: A Bool object indicating whether the objects are equal or not.
        """
        return self.serialize() == other.serialize()


@dataclass
class Float64:
    """
    someipy datatype representing a 64 bit floating type.
    """

    value: float = 0.0

    def __len__(self):
        """
        Return the length of the object.

        Returns:
            int: The length of the object, which is always 8.
        """
        return 8

    def serialize(self) -> bytes:
        """
        Serialize the value of the object into bytes using the big-endian byte order.

        Returns:
            bytes: The serialized value of the object.

        This method serializes the value of the object into bytes using the big-endian byte order. It expects the value to be a float. The serialized value is returned as a bytes object.
        """
        return struct.pack(">d", self.value)

    def deserialize(self, payload):
        """
        Deserialize the payload into the value of the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the value of the object. It expects the payload to be an 8-byte float in big-endian byte order. The deserialized value is assigned to the `value` attribute of the object. The deserialized object is then returned.
        """
        (self.value,) = struct.unpack(">d", payload)
        return self

    def __eq__(self, other) -> Bool:
        """
        Compare two objects for equality.

        This method compares the serialized representation of the current object and the other object to determine if they are equal. It returns a Bool object indicating whether the objects are equal or not.

        Parameters:
            other (Any): The object to compare with the current object.

        Returns:
            Bool: A Bool object indicating whether the objects are equal or not.
        """
        return self.serialize() == other.serialize()


def serialize(obj) -> bytes:
    """
    Serializes an object into bytes by iterating over its attributes, excluding those starting with double underscores or underscores.
    For each attribute, it calls the `serialize` method of the attribute and appends the returned bytes to the output.

    Parameters:
        obj (object): The object to be serialized.

    Returns:
        bytes: The serialized representation of the object as bytes.
    """
    ordered_items = [
        (name, value)
        for name, value in obj.__dict__.items()
        if not (name.startswith("__") or name.startswith("_"))
    ]
    output = bytes()
    for _, value in ordered_items:
        output += value.serialize()
    return output


class SomeIpPayload:
    """
    A base class for defining a custom SOME/IP payload ("structs"). It can be recursively nested, i.e. a SomeIpPayload object may contain other SomeIpPayload objects.
    """

    def __len__(self) -> int:
        """
        Return the length of the object.

        Returns:
            int: The length of the object.
        """
        payload_length = 0
        ordered_items = [
            (name, value)
            for name, value in self.__dict__.items()
            if not name.startswith("__")
        ]
        for _, value in ordered_items:
            payload_length += len(value)
        return payload_length

    def serialize(self) -> bytes:
        """
        Serialize the object into bytes to be transported on the wire.

        Returns:
            bytes: The serialized representation of the object.
        """
        return serialize(self)

    def deserialize(self, payload: bytes):
        """
        Deserialize the payload (on wire representation) into the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the object. It iterates over the attributes of the object, excluding those starting with "__". For each attribute, it calculates the length of the corresponding value and deserializes it using the `deserialize` method of the value object. The deserialized values are assigned back to the corresponding attributes of the object. Finally, the deserialized object is returned.
        """
        ordered_items = [
            (name, value)
            for name, value in self.__dict__.items()
            if not name.startswith("__")
        ]
        pos = 0

        for key, value in ordered_items:
            type_length = len(value)
            self.__dict__[key].deserialize(payload[pos : (pos + type_length)])
            pos += type_length
        return self


T = TypeVar("T")


class SomeIpFixedSizeArray(Generic[T]):
    """
    A datatype for a SOME/IP fixed size array. This type shall be used with someipy datatypes that support serialization and deserialization.
    """

    data: List[T]

    def __init__(self, class_reference: Type[T], size: int):
        """
        Initializes a new instance of the SomeIpFixedSizeArray class.

        Parameters:
            class_reference (Type[T]): The type of elements to be stored in the array.
            size (int): The number of elements in the array.

        Returns:
            None
        """
        self.data = [class_reference() for i in range(size)]

    def __eq__(self, other):
        """
        Compare two SomeIpFixedSizeArray objects for equality.

        This method compares the length (number of elements) of the current array and the other array to determine if they are equal. It also compares the length of the bytes representation of the arrays. Finally, it compares the content of all elements in the arrays to check if they are equal.

        Parameters:
            other (SomeIpFixedSizeArray): The object to compare with the current object.

        Returns:
            bool: True if the arrays are equal, False otherwise.
        """
        if isinstance(other, SomeIpFixedSizeArray):
            # Compare if the length (number of elements) of other array is the same
            if len(self.data) != len(other.data):
                return False

            # Compare if bytes length of other is the same
            if len(self) != len(other):
                return False

            # Compare if the content of all elements is the same
            for i in range(len(self.data)):
                if self.data[i] != other.data[i]:
                    return False
            return True

        return False

    def __len__(self) -> int:
        """
        Return the length of the object.

        This method calculates the length of the object based on the number of elements in the `data` list and the length of each element. If the `data` list is empty, it returns 0. Otherwise, it returns the product of the length of the `data` list and the length of the first element in the `data` list.

        Returns:
            int: The length of the object.
        """
        if len(self.data) == 0:
            return 0
        else:
            return len(self.data) * len(self.data[0])

    def serialize(self) -> bytes:
        """
        Serialize the object into bytes by iterating over its attributes, excluding those starting with double underscores or underscores.
        For each attribute, it calls the `serialize` method of the attribute and appends the returned bytes to the output.

        Returns:
            bytes: The serialized representation of the object as bytes.
        """
        result = bytes()
        for element in self.data:
            result += element.serialize()
        return result

    def deserialize(self, payload: bytes):
        """
        Deserialize the payload into the object.

        Args:
            payload (bytes): The payload to be deserialized.

        Returns:
            self: The deserialized object.

        This method deserializes the payload into the object. It iterates over the `data` list and for each element, it calls the `deserialize` method of that element, passing a slice of the payload corresponding to the element's length. The deserialized values are assigned back to the corresponding elements of the `data` list. If the `data` list is empty, the method returns immediately. Finally, the deserialized object is returned.
        """
        if len(self.data) == 0:
            return

        single_element_length = len(self.data[0])
        for i in range(len(self.data)):
            self.data[i].deserialize(
                payload[(i * single_element_length) : ((i + 1) * single_element_length)]
            )
        return self



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from typing import Tuple

from someipy._internal.someip_message import SomeIpMessage
from someipy.service import Service

from someipy._internal.tcp_client_manager import TcpClientManager, TcpClientProtocol
from someipy._internal.message_types import MessageType
from someipy._internal.return_codes import ReturnCode
from someipy._internal.someip_sd_builder import (
    build_stop_offer_service_sd_header,
    build_subscribe_eventgroup_ack_entry,
    build_offer_service_sd_header,
    build_subscribe_eventgroup_ack_sd_header,
)
from someipy._internal.someip_header import SomeIpHeader
from someipy._internal.someip_sd_header import (
    SdService,
    TransportLayerProtocol,
    SdEventGroupEntry,
    SdIPV4EndpointOption,
)
from someipy._internal.service_discovery_abcs import (
    ServiceDiscoveryObserver,
    ServiceDiscoverySender,
)
from someipy._internal.simple_timer import SimplePeriodicTimer
from someipy._internal.utils import (
    create_udp_socket,
    EndpointType,
    endpoint_to_str_int_tuple,
)
from someipy._internal.logging import get_logger
from someipy._internal.subscribers import Subscribers, EventGroupSubscriber
from someipy._internal.someip_endpoint import (
    SomeipEndpoint,
    TCPSomeipEndpoint,
    UDPSomeipEndpoint,
)

_logger_name = "server_service_instance"


class ServerServiceInstance(ServiceDiscoveryObserver):
    _service: Service
    _instance_id: int
    _endpoint: EndpointType
    _protocol: TransportLayerProtocol
    _someip_endpoint: SomeipEndpoint
    _ttl: int
    _sd_sender: ServiceDiscoverySender
    _cyclic_offer_delay_ms: int

    _subscribers: Subscribers
    _offer_timer: SimplePeriodicTimer

    _handler_tasks: set[asyncio.Task]
    _is_running: bool

    def __init__(
        self,
        service: Service,
        instance_id: int,
        endpoint: EndpointType,
        protocol: TransportLayerProtocol,
        someip_endpoint: SomeipEndpoint,
        ttl: int = 0,  # TTL used for SD Offer entries
        sd_sender=None,
        cyclic_offer_delay_ms=2000,
    ):
        self._service = service
        self._instance_id = instance_id
        self._endpoint = endpoint
        self._protocol = protocol
        self._someip_endpoint = someip_endpoint
        self._ttl = ttl
        self._sd_sender = sd_sender
        self._cyclic_offer_delay_ms = cyclic_offer_delay_ms

        self._subscribers = Subscribers()
        self._offer_timer = None

        self._handler_tasks = set()
        self._is_running = True

    def send_event(self, event_group_id: int, event_id: int, payload: bytes) -> None:
        """
        Sends an event to subscribers with the given event group ID, event ID, and payload.

        Args:
            event_group_id (int): The ID of the event group.
            event_id (int): The ID of the event.
            payload (bytes): The payload of the event. Can be manually crafter or serialized using someipy serialization.

        Returns:
            None: This function does not return anything.

        Note:
            - The client id and session id are set to 0x00 and 0x01 respectively.
            - The protocol version and interface version are set to 1.
        """

        self._subscribers.update()

        length = 8 + len(payload)
        someip_header = SomeIpHeader(
            service_id=self._service.id,
            method_id=event_id,
            length=length,
            client_id=0x00,  # TODO
            session_id=0x01,  # TODO
            protocol_version=1,  # TODO
            interface_version=1,  # TODO
            message_type=MessageType.NOTIFICATION.value,
            return_code=0x00,
        )

        for sub in self._subscribers.subscribers:
            # Check if the subscriber wants to receive the event group id
            if sub.eventgroup_id == event_group_id:
                get_logger(_logger_name).debug(
                    f"Send event for instance 0x{self._instance_id:04X}, service: 0x{self._service.id:04X} to {sub.endpoint[0]}:{sub.endpoint[1]}"
                )
                self._someip_endpoint.sendto(
                    someip_header.to_buffer() + payload,
                    endpoint_to_str_int_tuple(sub.endpoint),
                )

    async def _handle_method_call(self, method_handler, dst_addr, header_to_return):
        result = await method_handler
        header_to_return.message_type = result.message_type.value
        header_to_return.return_code = result.return_code.value
        payload_to_return = result.payload

        # Update length in header to the correct length
        header_to_return.length = 8 + len(payload_to_return)
        self._someip_endpoint.sendto(
            header_to_return.to_buffer() + payload_to_return, dst_addr
        )

    def someip_message_received(
        self, message: SomeIpMessage, addr: Tuple[str, int]
    ) -> None:
        """
        Handle a received Some/IP message, typically when a client uses an offered service.

        Args:
            message (SomeIpMessage): The received someip message.
            addr (Tuple[str, int]): The address of the sender consisting of IP address and source port.

        Returns:
            None: This function does not return anything.

        Raises:
            None: This function does not raise any exceptions.

        Note:
            - The protocol and interface version are not checked yet.
            - If the message type in the received header is not a request, a warning is logged.
        """

        if not self._is_running:
            return

        header = message.header
        payload_to_return = bytes()
        header_to_return = header

        def send_response():
            """Helper function to send out the buffer"""

            # Update length in header to the correct length
            header_to_return.length = 8 + len(payload_to_return)
            self._someip_endpoint.sendto(
                header_to_return.to_buffer() + payload_to_return, addr
            )

        if header.service_id != self._service.id:
            get_logger(_logger_name).warning(
                f"Unknown service ID received from {addr}: ID 0x{header.service_id:04X}"
            )
            header_to_return.message_type = MessageType.RESPONSE.value
            header_to_return.return_code = ReturnCode.E_UNKNOWN_SERVICE.value
            send_response()
            return

        if header.method_id not in self._service.methods.keys():
            get_logger(_logger_name).warning(
                f"Unknown method ID received from {addr}: ID 0x{header.method_id:04X}"
            )
            header_to_return.message_type = MessageType.RESPONSE.value
            header_to_return.return_code = ReturnCode.E_UNKNOWN_METHOD.value
            send_response()
            return

        # TODO: Test for protocol and interface version

        if (
            header.message_type == MessageType.REQUEST.value
            and header.return_code == 0x00
        ):
            method_handler = self._service.methods[header.method_id].method_handler
            coro = method_handler(message.payload, addr)

            # If a method is called, do it in a separate task to allow for asynchronous processing inside
            # method handlers
            new_task = asyncio.create_task(
                self._handle_method_call(coro, addr, header_to_return)
            )
            self._handler_tasks.add(new_task)
            new_task.add_done_callback(self._handler_tasks.discard)

        else:
            get_logger(_logger_name).warning(
                f"Unknown message type received from {addr}: Type 0x{header.message_type:04X}"
            )

    def handle_find_service(self):
        """
        Handle an SD find service entry. Not implemented yet.

        Args:
            None
        Returns:
            None
        """
        # TODO: implement SD behaviour and send back offer
        pass

    def handle_offer_service(self, _: SdService):
        """
        React on an SD offer entry. In a server instance no reaction is needed.

        Parameters:
            _ (SdService): The service update that was offered.

        Returns:
            None
        """
        # No reaction in a server instance needed
        pass

    def handle_stop_offer_service(self, _: SdService) -> None:
        # No reaction in a server instance needed
        pass

    def handle_subscribe_eventgroup(
        self,
        sd_event_group: SdEventGroupEntry,
        ipv4_endpoint_option: SdIPV4EndpointOption,
    ) -> None:
        """
        React on an SD subscribe eventgroup entry.

        Parameters:
            sd_event_group (SdEventGroupEntry): The subscribe eventgroup entry.
            ipv4_endpoint_option (SdIPV4EndpointOption): The IPv4 endpoint option.

        Returns:
            None

        """

        # From SD specification:
        # [PRS_SOMEIPSD_00829] When receiving a SubscribeEventgroupAck or Sub-
        # scribeEventgroupNack the Service ID, Instance ID, Eventgroup ID, and Major Ver-
        # sion shall match exactly to the corresponding SubscribeEventgroup Entry to identify
        # an Eventgroup of a Service Instance.
        # TODO: Enable major version check
        if sd_event_group.sd_entry.service_id != self._service.id:
            return
        if sd_event_group.sd_entry.instance_id != self._instance_id:
            return
        if sd_event_group.eventgroup_id not in self._service.eventgroupids:
            return

        if ipv4_endpoint_option.protocol != self._protocol:
            get_logger(_logger_name).warn(
                f"Subscribing a different protocol (TCP/UDP) than offered is not supported. Received subscribe for instance 0x{self._instance_id:04X}, service: 0x{self._service.id:04X} "
                "from {ipv4_endpoint_option.ipv4_address}/{ipv4_endpoint_option.port} with wrong protocol"
            )
            return

        (
            session_id,
            reboot_flag,
        ) = self._sd_sender.get_unicast_session_handler().update_session()

        get_logger(_logger_name).debug(
            f"Send Subscribe ACK for instance 0x{self._instance_id:04X}, service: 0x{self._service.id:04X}, TTL: {sd_event_group.sd_entry.ttl}"
        )
        ack_entry = build_subscribe_eventgroup_ack_entry(
            service_id=sd_event_group.sd_entry.service_id,
            instance_id=sd_event_group.sd_entry.instance_id,
            major_version=sd_event_group.sd_entry.major_version,
            ttl=sd_event_group.sd_entry.ttl,
            event_group_id=sd_event_group.eventgroup_id,
        )
        header_output = build_subscribe_eventgroup_ack_sd_header(
            entry=ack_entry, session_id=session_id, reboot_flag=reboot_flag
        )

        self._sd_sender.send_unicast(
            buffer=header_output.to_buffer(),
            dest_ip=ipv4_endpoint_option.ipv4_address,
        )

        self._subscribers.add_subscriber(
            EventGroupSubscriber(
                eventgroup_id=sd_event_group.eventgroup_id,
                endpoint=(
                    ipv4_endpoint_option.ipv4_address,
                    ipv4_endpoint_option.port,
                ),
                ttl=sd_event_group.sd_entry.ttl,
            )
        )

    def handle_subscribe_ack_eventgroup(self, _: SdEventGroupEntry) -> None:
        """
        React on an SD subscribe ACK event group entry. Shall not be received in a server instance. No reaction is needed in a server service instance.

        Args:
            _ (SdEventGroupEntry): The subscription ACK event group entry.

        Returns:
            None: This function does not return anything.
        """
        # Not needed for server instance
        pass

    def offer_timer_callback(self):
        """
        Callback function for the periodic offer timer.

        Parameters:
            None

        Returns:
            None
        """
        get_logger(_logger_name).debug(
            f"Offer service for instance 0x{self._instance_id:04X}, service: 0x{self._service.id:04X}, TTL: {self._ttl}, version: {self._service.major_version}.{self._service.minor_version}"
        )

        service_to_offer = SdService(
            service_id=self._service.id,
            instance_id=self._instance_id,
            major_version=1,
            minor_version=0,
            ttl=self._ttl,
            endpoint=self._endpoint,
            protocol=self._protocol,
        )
        self._sd_sender.offer_service(service_to_offer)

    def start_offer(self):
        """
        Starts sending periodic service discovery offer entries for the instance.

        This method creates a new `SimplePeriodicTimer` instance with the specified delay in seconds and assigns it to the `_offer_timer` attribute. The timer is then started by calling its `start` method.

        Parameters:
            None

        Returns:
            None
        """
        if self._offer_timer is None:
            self._offer_timer = SimplePeriodicTimer(
                self._cyclic_offer_delay_ms / 1000.0, self.offer_timer_callback
            )
            self._offer_timer.start()

    async def stop_offer(self):
        """
        Stop the offer for the service instance.  Stops the periodic send offer timer and sends out an SD message with a stop offer entry.

        Parameters:
            None

        Returns:
            None
        """
        get_logger(_logger_name).debug(
            f"Stop offer for instance 0x{self._instance_id:04X}, service: 0x{self._service.id:04X}"
        )

        self._subscribers.clear()

        if self._offer_timer is not None:
            self._offer_timer.stop()
            await self._offer_timer.task

        service_to_stop = SdService(
            service_id=self._service.id,
            instance_id=self._instance_id,
            major_version=1,
            minor_version=0,
            ttl=self._ttl,
            endpoint=self._endpoint,
            protocol=self._protocol,
        )
        (
            session_id,
            reboot_flag,
        ) = self._sd_sender.get_multicast_session_handler().update_session()
        sd_header = build_stop_offer_service_sd_header(
            [service_to_stop], session_id, reboot_flag
        )
        self._sd_sender.send_multicast(sd_header.to_buffer())

        # Stop processing incoming calls
        self._is_running = False

        # Cancel all running handler tasks
        for task in self._handler_tasks:
            task.cancel()

        # Wait for all tasks to be canceled
        await asyncio.gather(*self._handler_tasks)


async def construct_server_service_instance(
    service: Service,
    instance_id: int,
    endpoint: EndpointType,
    ttl,
    sd_sender: ServiceDiscoverySender,
    cyclic_offer_delay_ms=2000,
    protocol=TransportLayerProtocol.UDP,
) -> ServerServiceInstance:
    """
    Asynchronously constructs a ServerServiceInstance. Based on the given transport protocol, proper endpoints are setup before constructing the actual ServerServiceInstance.

    Args:
        service (Service): The service associated with the instance.
        instance_id (int): The ID of the instance.
        endpoint (EndpointType): The endpoint for the instance containing IP address and port.
        ttl (int, optional): The time-to-live for the instance used for service discovery offer entries. A value of 0 means that offer entries are valid for infinite time.
        sd_sender (Any, optional): The service discovery sender.
        cyclic_offer_delay_ms (int, optional): The delay in milliseconds for cyclic offers. Defaults to 2000.
        protocol (TransportLayerProtocol, optional): The transport layer protocol for the instance. Defaults to TransportLayerProtocol.UDP.

    Returns:
        ServerServiceInstance: The constructed ServerServiceInstance.

    Raises:
        None
    """
    if protocol == TransportLayerProtocol.UDP:
        loop = asyncio.get_running_loop()
        rcv_socket = create_udp_socket(str(endpoint[0]), endpoint[1])

        _, udp_endpoint = await loop.create_datagram_endpoint(
            lambda: UDPSomeipEndpoint(), sock=rcv_socket
        )

        server_instance = ServerServiceInstance(
            service,
            instance_id,
            endpoint,
            TransportLayerProtocol.UDP,
            udp_endpoint,
            ttl,
            sd_sender,
            cyclic_offer_delay_ms,
        )

        udp_endpoint.set_someip_callback(server_instance.someip_message_received)

        return server_instance

    elif protocol == TransportLayerProtocol.TCP:

        # Create a TcpClientManager, a TcpClientProtocol and a TCP server
        # The TcpClientProtocol handles incoming (or lost) connections and will (de)register them
        # in the TcpClientManager. The TcpClientProtocol also handles incoming data and will trigger
        # the callback in the TcpClientManager which will forward the callback to the ClientServiceInstance.
        tcp_client_manager = TcpClientManager()
        loop = asyncio.get_running_loop()
        server = await loop.create_server(
            lambda: TcpClientProtocol(client_manager=tcp_client_manager),
            str(endpoint[0]),
            endpoint[1],
        )

        tcp_someip_endpoint = TCPSomeipEndpoint(server, tcp_client_manager)

        server_instance = ServerServiceInstance(
            service,
            instance_id,
            endpoint,
            TransportLayerProtocol.TCP,
            tcp_someip_endpoint,
            ttl,
            sd_sender,
            cyclic_offer_delay_ms,
        )

        tcp_someip_endpoint.set_someip_callback(server_instance.someip_message_received)

        return server_instance



# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import List, Callable, Tuple, Dict

from someipy._internal.method_result import MethodResult


MethodHandler = Callable[[bytes, Tuple[str, int]], MethodResult]


@dataclass
class Method:
    """
    Class representing a SOME/IP method with a method id and a method handler.
    """

    id: int
    method_handler: MethodHandler

    def __eq__(self, __value: object) -> bool:
        return self.method_id == __value.method_id


@dataclass
class EventGroup:
    """
    Class representing a SOME/IP eventgroup with an eventgroup id and a list of event ids.
    """

    id: int
    event_ids: List[int]


@dataclass
class Service:
    """
    Class representing a SOME/IP service. A service has an id, major and minor version and 0 or more methods and/or eventgroups.
    """

    id: int
    major_version: int
    minor_version: int

    methods: Dict[int, Method]
    eventgroups: Dict[int, EventGroup]

    def __init__(self):
        self.id = 0
        self.major_version = 1
        self.minor_version = 0
        self.methods = dict()
        self.eventgroups = dict()

    @property
    def eventgroupids(self) -> List[int]:
        """
        Returns a list of event group IDs associated with the service.

        :return: A list of integers representing the event group IDs.
        """
        return list(self.eventgroups.keys())

    @property
    def methodids(self) -> List[int]:
        """
        Returns a list of method IDs associated with the object.

        :return: A list of integers representing the method IDs.
        :rtype: List[int]
        """
        return list(self.methods.keys())


class ServiceBuilder:
    """
    Class used to build a Service using a fluent API. Call the "with_" methods to add methods and event groups to the service. Call the build method to create the service.
    """

    def __init__(self):
        """
        Initializes a new ServiceBuilder instance.
        """
        self._service = Service()

    def with_service_id(self, id: int) -> "ServiceBuilder":
        """
        Sets the service id for the service to be built.

        :param id: The ID of the service.
        :type id: int
        :return: The ServiceBuilder instance.
        :rtype: ServiceBuilder
        """
        self._service.id = id
        return self

    def with_major_version(self, major_version: int) -> "ServiceBuilder":
        """
        Sets the major version for the service to be built.

        :param major_version: The major version of the service.
        :type major_version: int
        :return: The ServiceBuilder instance.
        :rtype: ServiceBuilder
        """
        self._service.major_version = major_version
        return self

    def with_minor_version(self, minor_version: int) -> "ServiceBuilder":
        """
        Sets the minor version for the service to be built.

        :param minor_version: The minor version of the service.
        :type minor_version: int
        :return: The ServiceBuilder instance.
        :rtype: ServiceBuilder
        """
        self._service.minor_version = minor_version
        return self

    def with_method(self, method: Method) -> "ServiceBuilder":
        """
        Adds a method to the service to be built.

        :param method: The method to be added.
        :type method: Method
        :return: The ServiceBuilder instance.
        :rtype: ServiceBuilder
        """
        if self._service.methods.get(method.id) is None:
            self._service.methods[method.id] = method
        return self

    def with_eventgroup(self, eventgroup: EventGroup) -> "ServiceBuilder":
        """
        Adds an event group to the service to be built.

        :param eventgroup: The event group to be added.
        :type eventgroup: EventGroup
        :return: The ServiceBuilder instance.
        :rtype: ServiceBuilder
        """
        if self._service.eventgroups.get(eventgroup.id) is None:
            self._service.eventgroups[eventgroup.id] = eventgroup
        return self

    def build(self) -> Service:
        """
        Builds and returns a Service instance.

        :return: The built Service instance.
        :rtype: Service
        """
        return self._service



import time
from someipy._internal.subscribers import EventGroupSubscriber, Subscribers

def test_eventgroupsubscriber_equal():
    e_1 = ("123", 5)
    e_2 = ("123", 5)
    e_3 = ("456", 4)

    s_1 = EventGroupSubscriber(1, e_1, 500)
    s_2 = EventGroupSubscriber(1, e_2, 600)
    assert s_1 == s_2

    s_2.endpoint = e_3
    assert s_1 != s_2

    s_2.endpoint = e_2
    assert s_1 == s_2

    s_2.eventgroup_id = 2
    assert s_1 != s_2

 
def test_add_remove_update():

    subscriber = EventGroupSubscriber(eventgroup_id=1, endpoint=("123", 1), ttl=1)

    subscribers = Subscribers()
    initial_ts = subscriber.last_ts_ms
    time.sleep(0.1)

    assert len(subscribers.subscribers) == 0
    subscribers.add_subscriber(subscriber)
    assert initial_ts != subscribers.subscribers[0].last_ts_ms
    assert len(subscribers.subscribers) == 1
    
    # Add same subscriber. Length shall not change, but timestamp shall be updated
    initial_ts = subscribers.subscribers[0].last_ts_ms
    time.sleep(0.2)
    subscribers.add_subscriber(subscriber)
    assert initial_ts != subscribers.subscribers[0].last_ts_ms
    assert len(subscribers.subscribers) == 1

    subscriber_2 = EventGroupSubscriber(eventgroup_id=2, endpoint=("123", 1), ttl=1)
    subscribers.add_subscriber(subscriber_2)
    assert len(subscribers.subscribers) == 2

    subscriber_3 = EventGroupSubscriber(eventgroup_id=2, endpoint=("123", 1), ttl=1)
    subscribers.remove_subscriber(subscriber_3)
    assert len(subscribers.subscribers) == 1

    subscriber_4 = EventGroupSubscriber(eventgroup_id=1, endpoint=("123", 1), ttl=1)
    subscribers.remove_subscriber(subscriber_4)
    assert len(subscribers.subscribers) == 0

    subscribers.add_subscriber(EventGroupSubscriber(eventgroup_id=1, endpoint=("123", 1), ttl=1))
    subscribers.add_subscriber(EventGroupSubscriber(eventgroup_id=2, endpoint=("123", 1), ttl=3))
    subscribers.add_subscriber(EventGroupSubscriber(eventgroup_id=3, endpoint=("123", 1), ttl=2))
    assert len(subscribers.subscribers) == 3
    time.sleep(1.1)
    subscribers.update()
    assert len(subscribers.subscribers) == 2
    time.sleep(3)
    subscribers.update()
    assert len(subscribers.subscribers) == 0


# Copyright (C) 2024 Christian H.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import ipaddress
import queue
from typing import Any, Iterable, Union, Tuple

from someipy._internal.someip_header import SomeIpHeader
from someipy._internal.someip_sd_builder import build_offer_service_sd_header
from someipy._internal.someip_sd_header import (
    SdEventGroupEntry,
    SdIPV4EndpointOption,
    SdService,
    SomeIpSdHeader,
)
from someipy._internal.someip_sd_extractors import (
    extract_offered_services,
    extract_subscribe_eventgroup_entries,
    extract_subscribe_ack_eventgroup_entries,
)
from someipy._internal.session_handler import SessionHandler
from someipy._internal.utils import (
    create_rcv_multicast_socket,
    create_udp_socket,
    DatagramAdapter,
)
from someipy._internal.service_discovery_abcs import (
    ServiceDiscoveryObserver,
    ServiceDiscoverySender,
    ServiceDiscoverySubject,
)
from someipy._internal.logging import get_logger

_logger_name = "service_discovery"


class ServiceDiscoveryProtocol(ServiceDiscoverySubject, ServiceDiscoverySender):
    """
    This class implements the ServiceDiscoverySubject and ServiceDiscoverySender
    interfaces. It is responsible for receiving and sending service discovery
    messages. Service instances are usually attached to objects of this class
    in order to receive service discovery updates, e.g. offer entries.
    """

    def __init__(self, multicast_ip: str, interface_ip: str, sd_port: int):
        """
        Initialize the ServiceDiscoveryProtocol instance.

        Args:
            multicast_ip (str): Multicast group IP address.
            interface_ip (str): IP address to be used as source address.
            sd_port (int): Service discovery port.
        """
        self.interface_ip = interface_ip
        self.multicast_ip = multicast_ip
        self.sd_port = sd_port

        self.attached_observers: Iterable[ServiceDiscoveryObserver] = []
        self.mcast_transport: asyncio.Transport = None
        self.unicast_transport: asyncio.Transport = None

        self.mcast_session_handler = SessionHandler()
        self.unicast_session_handler = SessionHandler()

        self.offer_service_queue = queue.Queue()

    def get_multicast_session_handler(self) -> SessionHandler:
        """
        Get the session handler for the multicast transport.

        Returns:
            SessionHandler: The session handler for service discovery multicast messages.
        """
        return self.mcast_session_handler

    def get_unicast_session_handler(self) -> SessionHandler:
        """
        Get the session handler for the unicast transport.

        Returns:
            SessionHandler: The session handler for service discovery unicast messages.
        """
        return self.unicast_session_handler

    def close(self):
        """
        Closes the transport sockets.
        """
        if self.mcast_transport is not None:
            self.mcast_transport.close()
        if self.unicast_transport is not None:
            self.unicast_transport.close()

    def detach(self, service_instance: ServiceDiscoveryObserver) -> None:
        """
        Detach a service instance. The service instance will no longer be notified about service discovery.

        Args:
            service_instance (ServiceDiscoveryObserver): The service instance to detach.
        """
        self.attached_observers.remove(service_instance)

    def attach(self, service_instance: ServiceDiscoveryObserver) -> None:
        """
        Attach a service instance to be notified about service discovery messages.

        Args:
            service_instance (ServiceDiscoveryObserver): The service instance to attach.
        """
        self.attached_observers.append(service_instance)

    def datagram_received(self, data: bytes, addr: Tuple[Union[str, Any], int]) -> None:
        """
        Handle a datagram received.

        Args:
            data (bytes): The received data.
            addr (Tuple[Union[str, Any], int]): The address of the sender.
        """
        # Ignore packets sent by the app itself
        if addr[0] == self.interface_ip:
            return

        # Ignore packets not sent to the SD port
        if addr[1] != self.sd_port:
            return

        someip_header = SomeIpHeader.from_buffer(data)
        if not someip_header.is_sd_header():
            return

        someip_sd_header = SomeIpSdHeader.from_buffer(data)

        for offered_service in extract_offered_services(someip_sd_header):
            self._handle_offered_service(offered_service)

        for (
            event_group_entry,
            ipv4_endpoint_option,
        ) in extract_subscribe_eventgroup_entries(someip_sd_header):
            self._handle_subscribe_eventgroup_entry(
                event_group_entry, ipv4_endpoint_option
            )

        for event_group_entry in extract_subscribe_ack_eventgroup_entries(
            someip_sd_header
        ):
            self._handle_subscribe_ack_eventgroup_entry(event_group_entry)

    def connection_lost(self, exc: Exception) -> None:
        """
        Handle connection lost.

        Args:
            exc (Exception): The exception that caused the connection loss.
        """
        pass

    def offer_service(self, service: SdService) -> None:
        # Put the service into a queue
        self.offer_service_queue.put(service)

        # Defer sending and wait for 20ms
        # This opens a time window of 20ms for other instances to offer services
        # which would be packed together into a single SD message
        asyncio.get_running_loop().call_later(0.02, self._sendout_offered_services)

    def _sendout_offered_services(self) -> None:
        services_to_offer = []
        while not self.offer_service_queue.empty():
            service = self.offer_service_queue.get()
            services_to_offer.append(service)
        if len(services_to_offer) > 0:
            (
                session_id,
                reboot_flag,
            ) = self.mcast_session_handler.update_session()

            sd_message = build_offer_service_sd_header(
                services_to_offer, session_id, reboot_flag
            )
            buffer = sd_message.to_buffer()
            self.send_multicast(buffer)

    def send_multicast(self, buffer: bytes) -> None:
        """
        Send a multicast message.

        Args:
            buffer (bytes): The message to send.
        """
        self.unicast_transport.sendto(buffer, (self.multicast_ip, self.sd_port))

    def send_unicast(self, buffer: bytes, dest_ip: ipaddress.IPv4Address) -> None:
        """
        Send a unicast message.

        Args:
            buffer (bytes): The message to send.
            dest_ip (ipaddress.IPv4Address): The destination IP address.
        """
        self.unicast_transport.sendto(buffer, (str(dest_ip), self.sd_port))

    def _handle_offered_service(self, offered_service: SdService) -> None:
        """
        Handle an offered service.

        Args:
            offered_service (SdService): The offered service.
        """
        get_logger(_logger_name).debug(
            f"Received offer for instance 0x{offered_service.instance_id:04X}, service 0x{offered_service.service_id:04X}"
        )
        for o in self.attached_observers:
            o.handle_offer_service(offered_service)

    def _handle_subscribe_eventgroup_entry(
        self,
        event_group_entry: SdEventGroupEntry,
        ipv4_endpoint_option: SdIPV4EndpointOption,
    ) -> None:
        """
        Handle a subscribe event group entry.

        Args:
            event_group_entry (SdEventGroupEntry): The event group entry.
            ipv4_endpoint_option (SdIPV4EndpointOption): The IPv4 endpoint option.
        """
        get_logger(_logger_name).debug(
            f"Received subscribe for instance 0x{event_group_entry.sd_entry.instance_id:04X}, service 0x{event_group_entry.sd_entry.service_id:04X}, eventgroup 0x{event_group_entry.eventgroup_id:04X}"
        )
        for o in self.attached_observers:
            o.handle_subscribe_eventgroup(event_group_entry, ipv4_endpoint_option)

    def _handle_subscribe_ack_eventgroup_entry(
        self, event_group_entry: SdEventGroupEntry
    ) -> None:
        """
        Handle a subscribe ACK event group entry.

        Args:
            event_group_entry (SdEventGroupEntry): The event group entry.
        """
        get_logger(_logger_name).debug(
            f"Received subscribe ACK for instance 0x{event_group_entry.sd_entry.instance_id:04X}, service 0x{event_group_entry.sd_entry.service_id:04X}, eventgroup 0x{event_group_entry.eventgroup_id:04X}"
        )
        for o in self.attached_observers:
            o.handle_subscribe_ack_eventgroup(event_group_entry)


async def construct_service_discovery(
    multicast_group_ip: str,
    sd_port: int,
    unicast_ip: str,
) -> ServiceDiscoveryProtocol:
    """
    Asynchronously constructs a ServiceDiscoveryProtocol instance.
    It opens one socket listening for multicast group address and
    one for listening on service discovery unicast traffic.
    The port is used both for multicast and unicast.

    :param multicast_group_ip: The service discovery multicast group IPv4 address.
    :type multicast_group_ip: str
    :param sd_port: The port number used for service discovery.
    :type sd_port: int
    :param unicast_ip: The IPv4 address used for unicast service discovery.
    :type unicast_ip: str
    :return: An instance of ServiceDiscoveryProtocol with transport protocols set up.
    :rtype: ServiceDiscoveryProtocol
    """
    sd = ServiceDiscoveryProtocol(
        multicast_group_ip,
        unicast_ip,
        sd_port,
    )

    loop = asyncio.get_running_loop()
    sd.mcast_transport, _ = await loop.create_datagram_endpoint(
        lambda: DatagramAdapter(target=sd),
        sock=create_rcv_multicast_socket(multicast_group_ip, sd_port, unicast_ip),
    )

    sd.unicast_transport, _ = await loop.create_datagram_endpoint(
        lambda: DatagramAdapter(target=sd),
        sock=create_udp_socket(unicast_ip, sd_port),
    )

    return sd



from .message_types import MessageType
from .return_codes import ReturnCode


class MethodResult:
    message_type: MessageType
    return_code: ReturnCode
    payload: bytes

    def __init__(self):
        self.message_type = MessageType.RESPONSE
        self.return_code = ReturnCode.E_OK
        self.payload = b""
